==> unicore-ucc6-5.0.0-1.src.tar.gz: ucc-core/src/main/java/de/fzj/unicore/ucc/helpers/Location.java <==
package de.fzj.unicore.ucc.helpers;

import java.util.regex.Matcher;
import java.util.regex.Pattern;

import org.w3.x2005.x08.addressing.EndpointReferenceType;

import de.fzj.unicore.uas.client.JobClient;
import de.fzj.unicore.uas.client.StorageClient;
import de.fzj.unicore.uas.client.TSSClient;
import de.fzj.unicore.ucc.MessageWriter;
import de.fzj.unicore.wsrflite.xmlbeans.client.IRegistryQuery;
import de.fzj.unicore.wsrflite.xmlbeans.client.RegistryClient;
import eu.unicore.security.util.client.IClientProperties;

/**
 * A local or remote file location. Can be specified in in the following ways
 * <ul>
 * 
 *   <li>either by a direct UNICORE 6 URI
 *   of the form <pre>PROTOCOL:ADDRESS#file</pre>
 * 
 *   <li>UNICORE 6 URI without protocol specification, 
 *   <pre>ADDRESS#file</pre>
 *  
 *   <li>by specifiying target system name and job id or Storage Name, 
 *   e.g. <pre>u6://MY-SITE/Home/filename</pre> <br/>
 *   In the latter form, the protocol can be specified by appending "?protocol=...", for example
 *   <pre>u6://MY-SITE/Home/filename?protocol=BFT</pre>
 * 
 *   <li>"Raw" locations (such as "ftp://..." or "http://..." URLs) are also supported
 * 
 * </ul>
 * 
 * If no protocol is given, BFT is used. <br/>
 * 
 * TODO honor the user's protocol preference if no protocol is given
 * 
 * @author schuller
 */
public class Location {

	protected boolean isLocal=false;
	protected boolean isRaw=false;

	protected String smsEpr;
	protected String protocol;
	protected String name;
	protected final MessageWriter msg;
	protected final IRegistryQuery registry;
	protected final IClientProperties sec;
	protected final boolean doResolve;

	/**
	 * Constructs a new Location
	 * 
	 * @param desc - the string describing the location
	 * @param registry - the {@link RegistryClient} pointing to the registry
	 * @param sec - security settings
	 * @param msg -  a {@link MessageWriter} instance
	 * @param doResolve - if <code>false</code>, no resolution of u6:// URLs will be performed
	 */
	public Location(String desc, IRegistryQuery registry, IClientProperties sec, MessageWriter msg, boolean doResolve){
		if(msg!=null){
			this.msg=msg;
		}
		else{
			this.msg=new DefaultMessageWriter();
		}
		this.doResolve=doResolve;
		this.registry=registry;
		this.sec=sec;
		if(parseDesc(desc)==false)throw new IllegalArgumentException("<"+desc+"> is not a valid location.");
	}

	/**
	 * Constructs a new Location
	 * 
	 * @param desc - the string describing the location
	 * @param registry - the {@link RegistryClient} pointing to the registry
	 * @param sec - security settings
	 * @param msg -  a {@link MessageWriter} instance
	 * @throws IllegalArgumentException in case the location is not valid or can not be resolved
	 */
	public Location(String desc, IRegistryQuery registry, IClientProperties sec, MessageWriter msg){
		this(desc,registry,sec,msg,true);
	}

	//pattern describing a U6 url PROTOCOL:https://<url>#/path
	final protected static String u6URLRE= "(([[\\w-]]+):)?([\\w-])+://(.)*#(.)*";
	final public static Pattern pattern = Pattern.compile(u6URLRE); 

	public static boolean isU6URL(String url){		
		Matcher m=pattern.matcher(url);
		return m.find();
	}

	final protected static String rawURLRE= "([\\w-])+:(.)*";
	final public static Pattern rawPattern = Pattern.compile(rawURLRE); 

	public static boolean isRawURL(String url){	
		boolean isU6=isU6URL(url);
		Matcher m=rawPattern.matcher(url);
		return !isU6 && m.find();
	}
	/**
	 * parses the supplied location. This can be either a 
	 * remote or local file
	 * 
	 * @param desc
	 * @return
	 */
	protected boolean parseDesc(String desc){
		//abstract u6:// URI?
		if(desc.startsWith("u6://") || desc.startsWith("unicore6://")){
			if(doResolve){
				boolean found=resolveLocation(desc);
				if(found){
					isLocal=false;
					isRaw=false;
				}
				return found;
			}
		}

		//check if it is an SMS UNICORE 6 URI
		if(isU6URL(desc)){
			String[] s1=desc.split("#");
			try{
				name=s1[s1.length-1];
				String s2[]=s1[0].split(":",2);
				if(s2[1].toLowerCase().startsWith("http")){
					protocol=s2[0];
					smsEpr=s2[1];
				}
				else{
					//no protocol, use BFT as fallback
					protocol="BFT";
					smsEpr=s1[0];
				}
				return true;
			}catch(Exception e){
				msg.error("Error: <"+desc+"> does not denote a valid UNICORE 6 URI",e);
				return false;
			}
		}
		//then check if it is a "raw" URL, i.e. does *not* have the form "PROTOTOL:scheme://..."
		if(isRawURL(desc)){
			smsEpr=desc;
			name=null;
			isRaw=true;
			return true;
		}

		//finally assume it is just a file path
		name=desc;
		isLocal=true;
		return true;

	}

	/**
	 * resolve a location by matching target system and storage names
	 * (inspired by the HiLA library's URI scheme).
	 * 
	 * The URL has the form
	 * 
	 * u6://TSS/STORAGE/FILE
	 * 
	 * or 
	 * 
	 * u6://TSS/jobid/FILE
	 * <br/>
	 * 
	 * (the file specification is optional)
	 * 
	 * Optionally, a protocol can be appended by "?protocol=..."<br/>
	 * 
	 * TODO honor preferred protocol
	 * 
	 * @param desc
	 * @return
	 */
	private boolean resolveLocation(String desc){
		if(sec==null)throw new IllegalArgumentException("Security properties may not be null.");
		if(registry==null)throw new IllegalArgumentException("Registry may not be null.");

		try{
			//first check if we have a protocol or set default protocol
			desc=resolveProtocol(desc);

			//now do the address
			String l=desc.split("://")[1];
			String[] pathSpec=l.split("/",3);
			String tssName=pathSpec[0];
			String location=null;
			if(pathSpec.length>1){
				location=pathSpec[1];
			}
			if(pathSpec.length>2){
				name=pathSpec[2];
				if(name==null || name.length()==0){
					name="/";
				}
			}

			boolean foundLocation=false;
			boolean fromCache=false;
			StorageClient sms=null;

			//check resource cache first
			ResourceCache cache=ResourceCache.getInstance();
			String cacheKey=tssName;
			//first check plain storage
			smsEpr=cache.getAs("locations", cacheKey, String.class);
			if(smsEpr==null && location!=null){
				//now try "SITE/StorageName"
				cacheKey=cacheKey+"/"+location;
				smsEpr=cache.getAs("locations", cacheKey, String.class);
			}
			else{
				getNameForStorage(l);
			}

			if(smsEpr!=null){
				foundLocation=true;
				fromCache=true;
			}
			if(smsEpr==null){
				//check TSS
				TSSClient tss=new TargetSystemLister(registry,sec,msg).getTSS(tssName);
				if(tss==null){
					//check if it is a shared storage registered in the registry
					foundLocation=checkStorage(tssName, l);
					if(foundLocation)cacheKey=tssName;
				}
				else{
					//find storage on TSS by name
					sms=new StorageLister(tss,sec,msg).getStorage(location);
					if(sms!=null){
						smsEpr=sms.getEPR().getAddress().getStringValue();
						foundLocation=true;
						cacheKey=tssName+"/"+location;
					}
					else{
						//check if it is a job on the TSS
						msg.verbose("Checking jobs on TSS <"+tssName+">.");
						//interpret location as a job ID
						for(EndpointReferenceType e: tss.getJobs()){
							try{
								msg.verbose("Check "+e.getAddress().getStringValue());
								if(e.getAddress().getStringValue().endsWith(location)){
									smsEpr=new JobClient(e.getAddress().getStringValue(),e,sec)
									.getResourcePropertiesDocument().getJobProperties()
									.getWorkingDirectoryReference().getAddress().getStringValue();
									foundLocation=true;
									cacheKey=tssName+"/"+location;
								}
							}catch(Throwable ex){
								msg.verbose("Error on access: "+ex.getMessage());
							}
						}
					}
				}
			}

			if(foundLocation){
				if(msg.isVerbose())msg.verbose("Resolved "+(fromCache?"cached":"")+" location "+desc+" as <"+smsEpr+">, file name <"+name+">");
				if(!fromCache)cache.put("locations", cacheKey, smsEpr);
			}
			else{
				msg.verbose("No matching location found!");
			}
			return foundLocation;
		}
		catch(Exception e){
			msg.error("Error resolving location.",e);
			return false;
		}
	}

	//resolve the protocol and return the descriptor without the protocol
	private String resolveProtocol(String desc){
		String[] protospec=desc.split("\\?");
		if(protospec.length==2){
			String p=protospec[1];
			desc=protospec[0];
			if(p.contains("protocol=")){
				protocol=p.split("protocol=")[1];
			}
			else{
				protocol="BFT";
			}
		}
		else{
			protocol="BFT";
		}
		return desc;
	}

	private boolean checkStorage(String storageName, String location){
		//check if it is a shared storage registered in the registry
		StorageClient sms=new StorageLister(registry,sec,msg).getStorage(storageName);
		if(sms!=null){
			smsEpr=sms.getEPR().getAddress().getStringValue();
			getNameForStorage(location);
			return true;
		}
		else{
			msg.verbose("No Target system or Storage named <"+storageName+"> found.");
		}
		return false;
	}

	//get the file name for a URL of the form "StorageName/filename"
	private void getNameForStorage(String location){
		String[] pathSpec=location.split("/",2);
		if(pathSpec.length>1){
			name=pathSpec[1];
		}
		if(name==null || name.length()==0){
			name="/";
		}
	}

	/**
	 * is this a local file?
	 *
	 * @return true if this is a local file
	 */
	public boolean isLocal() {
		return isLocal;
	}

	/**
	 * does this location denote a UNICORE 6 storage
	 *  
	 * @return <code>true</code> if it is a UNICORE URL
	 */
	public boolean isUnicoreURL() {
		return !isRaw && !isLocal;
	}

	/**
	 * does this location denote a "raw" storage, i.e. not a UNICORE 6 storage
	 * but something like a FTP URL?
	 *  
	 * @return true if it is a raw URL
	 */
	public boolean isRaw() {
		return isRaw;
	}

	public String getName() {
		return name;
	}

	public String getProtocol() {
		return protocol;
	}

	public void setProtocol(String protocol) {
		this.protocol=protocol;
	}

	public String getSmsEpr() {
		return smsEpr;
	}

	/**
	 * get this location as a UNICORE 6 URI, i.e.
	 * <code>protocol://SMS-URL#/path</code> <br/>
	 * for example
	 * <code>BFT://https://localhost:8080/DEMO-SITE/services/StorageManagement?res=default_storage#/test</code> <br/>
	 * @return
	 */
	public String getUnicore6URI() {
		if(isLocal() || smsEpr==null)return null;
		return protocol+":"+smsEpr+"#/"+name;
	}

	/**
	 * get the remote location this Location describes
	 * @return
	 */
	public String resolve(){
		if(isRaw)return smsEpr;
		return getUnicore6URI();
	}

}

==> unicore-ucc6-6.0.0-0.src.tar.gz: ucc-core/src/main/java/de/fzj/unicore/ucc/helpers/Location.java <==
package de.fzj.unicore.ucc.helpers;

import java.util.regex.Matcher;
import java.util.regex.Pattern;

import org.w3.x2005.x08.addressing.EndpointReferenceType;

import de.fzj.unicore.uas.client.JobClient;
import de.fzj.unicore.uas.client.StorageClient;
import de.fzj.unicore.uas.client.TSSClient;
import de.fzj.unicore.ucc.MessageWriter;
import de.fzj.unicore.wsrflite.xmlbeans.client.IRegistryQuery;
import de.fzj.unicore.wsrflite.xmlbeans.client.RegistryClient;
import eu.unicore.util.httpclient.IClientConfiguration;

/**
 * A local or remote file location. Can be specified in in the following ways
 * <ul>
 * 
 *   <li>either by a direct UNICORE 6 URI
 *   of the form <pre>PROTOCOL:ADDRESS#file</pre>
 * 
 *   <li>UNICORE 6 URI without protocol specification, 
 *   <pre>ADDRESS#file</pre>
 *  
 *   <li>by specifiying target system name and job id or Storage Name, 
 *   e.g. <pre>u6://MY-SITE/Home/filename</pre> <br/>
 *   In the latter form, the protocol can be specified by appending "?protocol=...", for example
 *   <pre>u6://MY-SITE/Home/filename?protocol=BFT</pre>
 * 
 *   <li>"Raw" locations (such as "ftp://..." or "http://..." URLs) are also supported
 * 
 * </ul>
 * 
 * If no protocol is given, BFT is used. <br/>
 * 
 * TODO honor the user's protocol preference if no protocol is given
 * 
 * @author schuller
 */
public class Location {

	protected boolean isLocal=false;
	protected boolean isRaw=false;

	protected String smsEpr;
	protected String protocol;
	protected String name;
	protected final MessageWriter msg;
	protected final IRegistryQuery registry;
	protected final IClientConfiguration sec;
	protected final boolean doResolve;

	/**
	 * Constructs a new Location
	 * 
	 * @param desc - the string describing the location
	 * @param registry - the {@link RegistryClient} pointing to the registry
	 * @param sec - security settings
	 * @param msg -  a {@link MessageWriter} instance
	 * @param doResolve - if <code>false</code>, no resolution of u6:// URLs will be performed
	 */
	public Location(String desc, IRegistryQuery registry, IClientConfiguration sec, MessageWriter msg, boolean doResolve){
		if(msg!=null){
			this.msg=msg;
		}
		else{
			this.msg=new DefaultMessageWriter();
		}
		this.doResolve=doResolve;
		this.registry=registry;
		this.sec=sec;
		if(parseDesc(desc)==false)throw new IllegalArgumentException("<"+desc+"> is not a valid location.");
	}

	/**
	 * Constructs a new Location
	 * 
	 * @param desc - the string describing the location
	 * @param registry - the {@link RegistryClient} pointing to the registry
	 * @param sec - security settings
	 * @param msg -  a {@link MessageWriter} instance
	 * @throws IllegalArgumentException in case the location is not valid or can not be resolved
	 */
	public Location(String desc, IRegistryQuery registry, IClientConfiguration sec, MessageWriter msg){
		this(desc,registry,sec,msg,true);
	}

	//pattern describing a U6 url PROTOCOL:https://<url>#/path
	final protected static String u6URLRE= "(([[\\w-]]+):)?([\\w-])+://(.)*#(.)*";
	final public static Pattern pattern = Pattern.compile(u6URLRE); 

	public static boolean isU6URL(String url){		
		Matcher m=pattern.matcher(url);
		return m.find();
	}

	final protected static String rawURLRE= "([\\w-])+:(.)*";
	final public static Pattern rawPattern = Pattern.compile(rawURLRE); 

	public static boolean isRawURL(String url){	
		boolean isU6=isU6URL(url);
		Matcher m=rawPattern.matcher(url);
		return !isU6 && m.find();
	}
	/**
	 * parses the supplied location. This can be either a 
	 * remote or local file
	 * 
	 * @param desc
	 * @return <code>true</code> if the location could be resolved
	 */
	protected boolean parseDesc(String desc){
		//abstract u6:// URI?
		if(desc.startsWith("u6://") || desc.startsWith("unicore6://")){
			if(doResolve){
				boolean found=resolveLocation(desc);
				if(found){
					isLocal=false;
					isRaw=false;
				}
				return found;
			}
		}

		//check if it is an SMS UNICORE 6 URI
		if(isU6URL(desc)){
			String[] s1=desc.split("#");
			try{
				name=s1[s1.length-1];
				String s2[]=s1[0].split(":",2);
				if(s2[1].toLowerCase().startsWith("http")){
					protocol=s2[0];
					smsEpr=s2[1];
				}
				else{
					//no protocol, use BFT as fallback
					protocol="BFT";
					smsEpr=s1[0];
				}
				return true;
			}catch(Exception e){
				msg.error("Error: <"+desc+"> does not denote a valid UNICORE 6 URI",e);
				return false;
			}
		}
		//then check if it is a "raw" URL, i.e. does *not* have the form "PROTOTOL:scheme://..."
		if(isRawURL(desc)){
			smsEpr=desc;
			name=null;
			isRaw=true;
			return true;
		}

		//finally assume it is just a file path
		name=desc;
		isLocal=true;
		return true;

	}

	/**
	 * resolve a location by matching target system and storage names
	 * (inspired by the HiLA library's URI scheme).
	 * 
	 * The URL has the form
	 * 
	 * u6://TSS/STORAGE/FILE
	 * 
	 * or 
	 * 
	 * u6://TSS/jobid/FILE
	 * <br/>
	 * 
	 * (the file specification is optional)
	 * 
	 * Optionally, a protocol can be appended by "?protocol=..."<br/>
	 * 
	 * TODO honor preferred protocol
	 * 
	 * @param desc
	 */
	private boolean resolveLocation(String desc){
		if(sec==null)throw new IllegalArgumentException("Security properties may not be null.");
		if(registry==null)throw new IllegalArgumentException("Registry may not be null.");

		try{
			//first check if we have a protocol or set default protocol
			desc=resolveProtocol(desc);

			//now do the address
			String l=desc.split("://")[1];
			String[] pathSpec=l.split("/",3);
			String tssName=pathSpec[0];
			String location=null;
			if(pathSpec.length>1){
				location=pathSpec[1];
			}
			if(pathSpec.length>2){
				name=pathSpec[2];
				if(name==null || name.length()==0){
					name="/";
				}
			}

			boolean foundLocation=false;
			boolean fromCache=false;
			StorageClient sms=null;

			//check resource cache first
			ResourceCache cache=ResourceCache.getInstance();
			String cacheKey=tssName;
			//first check plain storage
			smsEpr=cache.getAs("locations", cacheKey, String.class);
			if(smsEpr==null && location!=null){
				//now try "SITE/StorageName"
				cacheKey=cacheKey+"/"+location;
				smsEpr=cache.getAs("locations", cacheKey, String.class);
			}
			else{
				getNameForStorage(l);
			}

			if(smsEpr!=null){
				foundLocation=true;
				fromCache=true;
			}
			if(smsEpr==null){
				//check TSS
				TSSClient tss=new TargetSystemLister(registry,sec,msg).getTSS(tssName);
				if(tss==null){
					//check if it is a shared storage registered in the registry
					foundLocation=checkStorage(tssName, l);
					if(foundLocation)cacheKey=tssName;
				}
				else{
					//find storage on TSS by name
					sms=new StorageLister(tss,sec,msg).getStorage(location);
					if(sms!=null){
						smsEpr=sms.getEPR().getAddress().getStringValue();
						foundLocation=true;
						cacheKey=tssName+"/"+location;
					}
					else{
						//check if it is a job on the TSS
						msg.verbose("Checking jobs on TSS <"+tssName+">.");
						//interpret location as a job ID
						for(EndpointReferenceType e: tss.getJobs()){
							try{
								msg.verbose("Check "+e.getAddress().getStringValue());
								if(e.getAddress().getStringValue().endsWith(location)){
									smsEpr=new JobClient(e.getAddress().getStringValue(),e,sec)
									.getResourcePropertiesDocument().getJobProperties()
									.getWorkingDirectoryReference().getAddress().getStringValue();
									foundLocation=true;
									cacheKey=tssName+"/"+location;
								}
							}catch(Throwable ex){
								msg.verbose("Error on access: "+ex.getMessage());
							}
						}
					}
				}
			}

			if(foundLocation){
				if(msg.isVerbose())msg.verbose("Resolved "+(fromCache?"cached":"")+" location "+desc+" as <"+smsEpr+">, file name <"+name+">");
				if(!fromCache)cache.put("locations", cacheKey, smsEpr);
			}
			else{
				msg.verbose("No matching location found!");
			}
			return foundLocation;
		}
		catch(Exception e){
			msg.error("Error resolving location.",e);
			return false;
		}
	}

	//resolve the protocol and return the descriptor without the protocol
	private String resolveProtocol(String desc){
		String[] protospec=desc.split("\\?");
		if(protospec.length==2){
			String p=protospec[1];
			desc=protospec[0];
			if(p.contains("protocol=")){
				protocol=p.split("protocol=")[1];
			}
			else{
				protocol="BFT";
			}
		}
		else{
			protocol="BFT";
		}
		return desc;
	}

	private boolean checkStorage(String storageName, String location){
		//check if it is a shared storage registered in the registry
		StorageClient sms=new StorageLister(registry,sec,msg).getStorage(storageName);
		if(sms!=null){
			smsEpr=sms.getEPR().getAddress().getStringValue();
			getNameForStorage(location);
			return true;
		}
		else{
			msg.verbose("No Target system or Storage named <"+storageName+"> found.");
		}
		return false;
	}

	//get the file name for a URL of the form "StorageName/filename"
	private void getNameForStorage(String location){
		String[] pathSpec=location.split("/",2);
		if(pathSpec.length>1){
			name=pathSpec[1];
		}
		if(name==null || name.length()==0){
			name="/";
		}
	}

	/**
	 * is this a local file?
	 *
	 * @return true if this is a local file
	 */
	public boolean isLocal() {
		return isLocal;
	}

	/**
	 * does this location denote a UNICORE 6 storage
	 *  
	 * @return <code>true</code> if it is a UNICORE URL
	 */
	public boolean isUnicoreURL() {
		return !isRaw && !isLocal;
	}

	/**
	 * does this location denote a "raw" storage, i.e. not a UNICORE 6 storage
	 * but something like a FTP URL?
	 *  
	 * @return true if it is a raw URL
	 */
	public boolean isRaw() {
		return isRaw;
	}

	public String getName() {
		return name;
	}

	public String getProtocol() {
		return protocol;
	}

	public void setProtocol(String protocol) {
		this.protocol=protocol;
	}

	public String getSmsEpr() {
		return smsEpr;
	}

	/**
	 * get this location as a UNICORE 6 URI, i.e.
	 * <code>protocol://SMS-URL#/path</code> <br/>
	 * for example
	 * <code>BFT://https://localhost:8080/DEMO-SITE/services/StorageManagement?res=default_storage#/test</code> <br/>
	 */
	public String getUnicore6URI() {
		if(isLocal() || smsEpr==null)return null;
		return protocol+":"+smsEpr+"#/"+name;
	}

	/**
	 * get the remote location this Location describes
	 */
	public String resolve(){
		if(isRaw)return smsEpr;
		return getUnicore6URI();
	}

}
